#include "bms_channel.h"

#include "bms_exception.h"


// -- BmsChannelBuffer ---------------------------------------------------
BmsChannelBuffer::BmsChannelBuffer( void ) :
BmsBuffer( BmsChannelBuffer::DEFAULT_LENGTH )
{
}

void
BmsChannelBuffer::ExtendArrayOver( unsigned int pos )
{
  unsigned int extended_length = this->GetLength();
  while ( extended_length <= pos ) {
    extended_length *= 2;
  }
  if ( extended_length > this->GetLength() ) {
    this->ExtendTo( extended_length );
  }
}

BmsWord&
BmsChannelBuffer::operator []( unsigned int pos )
{
  this->ExtendArrayOver( pos );
  return this->BmsBuffer::operator []( pos );
}


void
BmsChannelBuffer::Merge( const BmsBuffer& buffer )
{
  this->ExtendArrayOver( buffer.GetLength() );
  this->BmsBuffer::Merge( buffer );
}

void
BmsChannelBuffer::Merge( unsigned int start, const BmsBuffer& buffer )
{
  this->ExtendArrayOver( start + buffer.GetLength() );
  this->BmsBuffer::Merge( start, buffer );
}


int
BmsChannelBuffer::GetObjectExistsMaxPosition( unsigned int start ) const
{
  int pos = -1;
  for ( unsigned int i = start + 1; i < this->GetLength(); ++i ) {
    if ( this->At( i ) != BmsWord::MIN ) {
      pos = i;
    }
  }
  return pos;
}


void
BmsChannelBuffer::MultiplyBarDivisionCount( unsigned int multiplier )
{
  this->MagnifyBy( multiplier );
}

// -- BmsChannel ---------------------------------------------------------
BmsChannel::BmsChannel( BmsWord channel_number ) :
channel_number_( channel_number ),
buffers_( 0 )
{
}

BmsChannel::~BmsChannel()
{
  for ( unsigned int i = 0; i < buffers_.size(); ++i ) {
    delete buffers_[i];
  }
}


void
BmsChannel::SetChannelNumber( BmsWord channel_number )
{
  channel_number_ = channel_number;
}

BmsWord
BmsChannel::GetChannelNumber() const
{
  return channel_number_;
}


unsigned int
BmsChannel::GetBufferCount( void ) const
{
  return buffers_.size();
}


BmsChannelBuffer&
BmsChannel::MakeNewBuffer( void )
{
  buffers_.push_back( new BmsChannelBuffer() );
  return *buffers_.back();
}


BmsChannelBuffer&
BmsChannel::GetBuffer( unsigned int pos )
{
  while ( pos >= buffers_.size() ) {
    this->MakeNewBuffer();
  }
  return *buffers_[pos];
}

BmsChannelBuffer&
BmsChannel::operator []( unsigned int pos )
{
  return this->GetBuffer( pos );
}

BmsChannel::Iterator
BmsChannel::Begin( void )
{
  return buffers_.begin();
}

BmsChannel::Iterator
BmsChannel::End( void )
{
  return buffers_.end();
}

BmsChannel::ConstIterator
BmsChannel::Begin( void ) const
{
  return buffers_.begin();
}

BmsChannel::ConstIterator
BmsChannel::End( void ) const
{
  return buffers_.end();
}


bool
BmsChannel::Contains( BmsWord word ) const
{
  for ( ConstIterator it = this->Begin(); it != this->End(); ++it ) {
    if ( (*it)->Contains( word ) ) {
      return true;
    }
  }
  return false;
}



int
BmsChannel::GetObjectExistsMaxPosition( unsigned int start ) const
{
  int pos = -1;
  for ( ConstIterator it = this->Begin(); it != this->End(); ++it ) {
    int tmp = (*it)->GetObjectExistsMaxPosition( static_cast<int>( start ) > pos ? static_cast<int>( start ) : pos );
    if ( tmp > pos ) {
      pos = tmp;
    }
  }
  return pos;
}


void
BmsChannel::MultiplyBarDivisionCount( unsigned int multiplier )
{
  for ( ConstIterator it = this->Begin(); it != this->End(); ++it ) {
    (*it)->MultiplyBarDivisionCount( multiplier );
  }
}


bool
BmsChannel::IsWavChannel( void ) const
{
  int c = channel_number_.ToInteger();
  return (
    // 01, 11 -- 1Z, 21 -- 2Z, 31 -- 3Z, 41 -- 4Z, 51 -- 5Z, 61 -- 6Z
    c == 1 ||
    37  <= c && c <= 71  ||
    73  <= c && c <= 107 ||
    109 <= c && c <= 143 ||
    145 <= c && c <= 179 ||
    181 <= c && c <= 215 ||
    217 <= c && c <= 251 );
}

bool
BmsChannel::IsShoudPlayWavChannel( void ) const
{
  int c = channel_number_.ToInteger();
  return (
    // 01, 11 -- 1Z, 21 -- 2Z, 51 -- 5Z, 61 -- 6Z
    c == 1 ||
    37  <= c && c <= 71  ||
    73  <= c && c <= 107 ||
    181 <= c && c <= 215 ||
    217 <= c && c <= 251 );
}

bool
BmsChannel::IsBmpChannel( void ) const
{
  switch ( channel_number_.ToInteger() ) {
    // 04 06 07 0A A5
  case 4:
  case 6:
  case 7:
  case 10:
  case 365:
    return true;

  default:
    return false;
  }
}

bool
BmsChannel::IsBpmChangeChannel( void ) const
{
  return channel_number_.ToInteger() == 3;
}

bool
BmsChannel::IsInvisibilityObjectChannel( void ) const
{
  int c = channel_number_.ToInteger();
  return (
    // 31 -- 3Z, 41 -- 4Z
    109 <= c && c <= 143 ||
    145 <= c && c <= 179 );
}

bool
BmsChannel::IsExtendedBpmChangeChannel( void ) const
{
  return channel_number_.ToInteger() == 8;
}

bool
BmsChannel::IsLongNoteChannel( void ) const
{
  int c = channel_number_.ToInteger();
  return (
    // 51 -- 5Z, 61 -- 6Z
    181 <= c && c <= 215 ||
    217 <= c && c <= 251 );
  
}

bool
BmsChannel::IsStopSequenceChannel( void ) const
{
  return channel_number_.ToInteger() == 9;
}


// -- BmsChannelManager --------------------------------------------------
BmsChannelManager::BmsChannelManager( void ) :
channels_()
{
}

BmsChannelManager::~BmsChannelManager()
{
  for ( std::map<BmsWord, BmsChannel*>::iterator it = channels_.begin(); it != channels_.end(); ++it ) {
    delete it->second;
  }
}

unsigned int
BmsChannelManager::GetCount( void ) const
{
  return channels_.size();
}

BmsChannel&
BmsChannelManager::GetChannel( BmsWord channel_number )
{
  if ( channels_.find( channel_number ) == channels_.end() ) {
    channels_[channel_number] = new BmsChannel( channel_number );
  }
  return *channels_[channel_number];
}

BmsChannel&
BmsChannelManager::operator []( BmsWord channel_number )
{
  return this->GetChannel( channel_number );
}


bool
BmsChannelManager::Contains( BmsWord word ) const
{
  for ( ConstIterator it = this->Begin(); it != this->End(); ++it ) {
    if ( it->second->Contains( word ) ) {
      return true;
    }
  }
  return false;
}

bool
BmsChannelManager::Contains( BmsWord word, BmsChannelConditionJudgeFunction func ) const
{
  for ( ConstIterator it = this->Begin(); it != this->End(); ++it ) {
    if ( (it->second->*func)() && it->second->Contains( word ) ) {
      return true;
    }
  }
  return false;
}


void
BmsChannelManager::DeleteChannel( BmsWord channel_number )
{
  std::map<BmsWord, BmsChannel*>::iterator it = channels_.find( channel_number );
  if ( it != channels_.end() ) {
    delete it->second;
    channels_.erase( it );
  }
}


BmsChannelManager::Iterator
BmsChannelManager::Begin( void )
{
  return channels_.begin();
}

BmsChannelManager::Iterator
BmsChannelManager::End( void )
{
  return channels_.end();
}

BmsChannelManager::ConstIterator
BmsChannelManager::Begin( void ) const
{
  return channels_.begin();
}

BmsChannelManager::ConstIterator
BmsChannelManager::End( void ) const
{
  return channels_.end();
}


int
BmsChannelManager::GetObjectExistsMaxPosition( void ) const
{
  int pos = -1;
  for ( ConstIterator it = this->Begin(); it != this->End(); ++it ) {
    int tmp = it->second->GetObjectExistsMaxPosition( pos > 0 ? pos : 0 );
    if ( tmp > pos ) {
      pos = tmp;
    }
  }
  return pos;
}


void
BmsChannelManager::MultiplyBarDivisionCount( unsigned int multiplier )
{
  for ( ConstIterator it = this->Begin(); it != this->End(); ++it ) {
    it->second->MultiplyBarDivisionCount( multiplier );
  }
}
